from typing import Callable, TypedDict, Any, List
from uuid import uuid4
from bullmq.scripts import Scripts
from bullmq.redis_connection import RedisConnection
from bullmq.event_emitter import EventEmitter
from bullmq.job import Job
from bullmq.timer import Timer

import asyncio
import traceback
import time


class WorkerOptions(TypedDict, total=False):
    autorun: bool
    """
    Condition to start processor at instance creation

    @default true
    """

    concurrency: int
    """
    Amount of jobs that a single worker is allowed to work on
    in parallel.

    @default 1
    @see https://docs.bullmq.io/guide/workers/concurrency
    """

    maxStalledCount: int
    """
    Amount of times a job can be recovered from a stalled state
    to the `wait` state. If this is exceeded, the job is moved
    to `failed`.

    @default 1
    """

    stalledInterval: int
    """
    Number of milliseconds between stallness checks.

    @default 30000
    """

    lockDuration: int
    """
    Duration of the lock for the job in milliseconds. The lock represents that
    a worker is processing the job. If the lock is lost, the job will be eventually
    be picked up by the stalled checker and move back to wait so that another worker
    can process it again.

    @default 30000
    """

    prefix: str
    """
    Prefix for all queue keys.
    """

    connection: dict[str, Any]
    """
    Options for connecting to a Redis instance.
    """


class Worker(EventEmitter):
    def __init__(self, name: str, processor: Callable[[Job, str], asyncio.Future], opts: WorkerOptions = {}):
        super().__init__()
        self.name = name
        self.processor = processor
        self.opts = {
            "concurrency": opts.get("concurrency", 1),
            "lockDuration": opts.get("lockDuration", 30000),
            "maxStalledCount": opts.get("maxStalledCount", 1),
            "stalledInterval": opts.get("stalledInterval", 30000),
        }
        redis_opts = opts.get("connection", {})
        self.redisConnection = RedisConnection(redis_opts)
        self.blockingRedisConnection = RedisConnection(redis_opts)
        self.client = self.redisConnection.conn
        self.bclient = self.blockingRedisConnection.conn
        self.scripts = Scripts(opts.get("prefix", "bull"), name, self.client)
        self.closing = False
        self.forceClosing = False
        self.closed = False
        self.running = False
        self.processing = set()
        self.jobs = set()

        if opts.get("autorun", True):
            asyncio.ensure_future(self.run())

    async def run(self):
        if self.running:
            raise Exception("Worker is already running")

        self.timer = Timer(
            (self.opts.get("lockDuration") / 2) / 1000, self.extendLocks)
        self.stalledCheckTimer = Timer(self.opts.get(
            "stalledInterval") / 1000, self.runStalledJobsCheck)
        self.running = True
        jobs = []

        token = uuid4().hex

        while not self.closed:
            if len(jobs) == 0 and len(self.processing) < self.opts.get("concurrency") and not self.closing:
                waiting_job = asyncio.ensure_future(self.getNextJob(token))
                self.processing.add(waiting_job)

            if len(jobs) > 0:
                jobs_to_process = [self.processJob(job, token) for job in jobs]
                processing_jobs = [asyncio.ensure_future(
                    j) for j in jobs_to_process]
                self.processing.update(processing_jobs)

            try:
                jobs, pending = await getCompleted(self.processing)

                self.processing = pending

                if (len(jobs) == 0 or len(self.processing) == 0) and self.closing:
                    # We are done processing so we can close the queue
                    break

            except Exception as e:
                # This should never happen or we will have an endless loop
                print("ERROR:", e)
                traceback.print_exc()
                return

        self.running = False
        self.timer.stop()
        self.stalledCheckTimer.stop()

    async def getNextJob(self, token: str):
        """
        Returns a promise that resolves to the next job in queue.
        @param token: worker token to be assigned to retrieved job
        @returns a Job or undefined if no job was available in the queue.
        """
        # First try to move a job from the waiting list to the active list
        result = await self.scripts.moveToActive(token, self.opts)
        job = None
        job_id = None
        delay_until = None
        if result:
            job, job_id = result

        # If there are no jobs in the waiting list we keep waiting with BRPOPLPUSH
        if job is None:
            timeout = min(delay_until - int(time.time() * 1000)
                          if delay_until else 5000, 5000) / 1000
            job_id = await self.bclient.brpoplpush(self.scripts.keys["wait"], self.scripts.keys["active"], timeout)
            if job_id:
                job, job_id = await self.scripts.moveToActive(token, self.opts, job_id)

        if job and job_id:
            return Job.fromJSON(self.client, job, job_id)

    async def processJob(self, job: Job, token: str):
        try:
            self.jobs.add((job, token))
            result = await self.processor(job, token)
            if not self.forceClosing:
                await self.scripts.moveToCompleted(job, result, job.opts.get("removeOnComplete", False), token, self.opts, fetchNext=not self.closing)
            self.emit("completed", job, result)
        except Exception as err:
            try:
                print("Error processing job", err)
                stacktrace = traceback.format_exc()

                if not self.forceClosing:
                    await self.scripts.moveToFailed(job, str(err), job.opts.get("removeOnFail", False), token, self.opts, fetchNext=not self.closing)

                # TODO: Store the stacktrace in the job

                self.emit("failed", job, err)
            except Exception as err:
                print("Error moving job to failed", err)
                self.emit("error", err, job)
        finally:
            self.jobs.remove((job, token))

    async def extendLocks(self):
        # Renew all the locks for the jobs that are still active
        try:
            multi = self.client.pipeline()
            for job, token in self.jobs:
                await self.scripts.extendLock(job.id, token, self.opts.get("lockDuration"), multi)
            result = await multi.execute()

            # result includes an object with locks that may not have been renewed.
            # We should emit an error for each of those jobs.
            #    for jobId, err in result.items():
            #    self.emit("error", "could not renew lock for job " + jobId)

        except Exception as e:
            print("Error renewing locks", e)
            traceback.print_exc()

    async def runStalledJobsCheck(self):
        try:
            failed, stalled = await self.scripts.moveStalledJobsToWait(self.opts.get("maxStalledCount"), self.opts.get("stalledInterval"))
            for jobId in failed:
                self.emit("failed", jobId,
                          "job stalled more than allowable limit")
            for jobId in stalled:
                self.emit("stalled", jobId)

        except Exception as e:
            print("Error checking stalled jobs", e)
            self.emit('error', e)

    async def close(self, force: bool = False):
        """
        Close the worker
        """
        if force:
            self.forceClosing = True

        self.closing = True

        await self.blockingRedisConnection.close()
        await self.redisConnection.close()


async def getCompleted(task_set: set) -> tuple[list[Job], list]:
    job_set, pending = await asyncio.wait(task_set, return_when=asyncio.FIRST_COMPLETED)
    jobs = [extract_result(job_task) for job_task in job_set]
    # we filter `None` out to remove:
    # a) an empty 'completed jobs' list; and
    # b) a failed extract_result
    jobs = list(filter(lambda j: j is not None, jobs))
    return jobs, pending


def extract_result(job_task):
    try:
        return job_task.result()
    except Exception as e:
        # lets use a simple-but-effective error handling:
        # print error message and ignore the job
        print("ERROR:", e)
        traceback.print_exc()



import redis.asyncio as redis


class RedisConnection:
    """
    RedisConnection class
    """

    def __init__(self, redisOpts: dict = {}):
        host = redisOpts.get("host") or "localhost"
        port = redisOpts.get("port") or 6379
        db = redisOpts.get("db") or 0
        password = redisOpts.get("password") or None

        self.conn = redis.Redis(
            host=host, port=port, db=db, password=password, decode_responses=True)

    def disconnect(self):
        """
        Disconnect from Redis
        """
        return self.conn.disconnect()

    def close(self):
        """
        Close the connection
        """
        return self.conn.close()



from setuptools import setup

# To use a consistent encoding
from codecs import open
from os import path

# Get the long description from the README file
with open(path.join(".", 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='bullmq',
    version='0.3.0',    
    description='BullMQ for Python',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://bullmq.io',
    author='Taskforce.sh Inc.',
    author_email='manast@taskforce.sh',
    license='MIT',
    packages=['bullmq'],
    package_data={'bullmq': ['commands/*.lua']},
    install_requires=['redis',
                      'msgpack',            
                      ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',  
        'Operating System :: POSIX :: Linux',        
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)



from redis import Redis
from typing import List, Any, TypedDict

import json
import time


optsDecodeMap = {
    'fpof': 'failParentOnFailure',
    'kl': 'keepLogs',
}

optsEncodeMap = {v: k for k, v in optsDecodeMap.items()}


class KeepJobs(TypedDict, total=False):
    """
    Specify which jobs to keep after finishing. If both age and count are
    specified, then the jobs kept will be the ones that satisfies both
    properties.
    """

    age: int
    """
    Maximum age in seconds for job to be kept.
    """

    count: int
    """
    Maximum count of jobs to be kept.
    """


class JobOptions(TypedDict, total=False):
    jobId: str
    """
    Override the job ID - by default, the job ID is a unique
    integer, but you can use this setting to override it.

    If you use this option, it is up to you to ensure the
    jobId is unique. If you attempt to add a job with an id that
    already exists, it will not be added.
    """

    timestamp: int
    """
    Timestamp when the job was created.

    @defaultValue round(time.time() * 1000)
    """

    delay: int
    """
    An amount of milliseconds to wait until this job can be processed.
    Note that for accurate delays, worker and producers
    should have their clocks synchronized.

    @defaultValue 0
    """

    attempts: int
    """
    The total number of attempts to try the job until it completes.

    @defaultValue 0
    """

    removeOnComplete: bool | int | KeepJobs
    """
    If true, removes the job when it successfully completes
    When given a number, it specifies the maximum amount of
    jobs to keep, or you can provide an object specifying max
    age and/or count to keep. It overrides whatever setting is used in the worker.
    Default behavior is to keep the job in the completed set.
    """

    removeOnFail: bool | int | KeepJobs
    """
    If true, removes the job when it fails after all attempts.
    When given a number, it specifies the maximum amount of
    jobs to keep, or you can provide an object specifying max
    age and/or count to keep. It overrides whatever setting is used in the worker.
    Default behavior is to keep the job in the failed set.
    """


class Job:
    """
    This class represents a Job in the queue. Normally job are implicitly created when
    you add a job to the queue with methods such as Queue.addJob( ... )

    A Job instance is also passed to the Worker's process function.
    """

    def __init__(self, client: Redis, name: str, data: Any, opts: JobOptions = {}):
        self.name = name
        self.id = opts.get("jobId", None)
        self.progress = 0
        self.timestamp = opts.get("timestamp", round(time.time() * 1000))
        self.opts = opts
        self.delay = opts.get("delay", 0)
        self.attempts = opts.get("attempts", 1)
        self.attemptsMade = 0
        self.data = data
        self.removeOnComplete = opts.get("removeOnComplete", True)
        self.removeOnFail = opts.get("removeOnFail", False)
        self.processedOn = 0
        self.finishedOn = 0
        self.returnvalue = None
        self.failedReason = None
        self.repeatJobKey = None
        self.stacktrace: List[str] = []


def fromJSON(client: Redis, rawData: dict, jobId: str | None = None):
    """
    Instantiates a Job from a JobJsonRaw object (coming from a deserialized JSON object)

    @param queue: the queue where the job belongs to.
    @param json: the plain object containing the job.
    @param jobId: an optional job id (overrides the id coming from the JSON object)
    """
    data = json.loads(rawData.get("data", '{}'))
    opts = optsFromJSON(json.loads(rawData.get("opts", '{}')))

    job = Job(client, rawData.get("name"), data, opts)
    job.id = jobId or rawData.get("id", b'').decode("utf-8")

    job.progress = json.loads(rawData.get("progress",  '0'))
    job.delay = int(rawData.get("delay", "0"))
    job.timestamp = int(rawData.get("timestamp", "0"))

    if rawData.get("finishedOn"):
        job.finishedOn = int(rawData.get("finishedOn"))

    if rawData.get("processedOn"):
        job.processedOn = int(rawData.get("processedOn"))

    if rawData.get("rjk"):
        job.repeatJobKey = rawData.get("rjk")

    job.failedReason = rawData.get("failedReason")
    job.attemptsMade = int(rawData.get("attemptsMade", "0"))

    returnvalue = rawData.get("returnvalue")
    if type(returnvalue) == str:
        job.returnvalue = getReturnValue(returnvalue)

    job.stacktrace = json.loads(rawData.get("stacktrace", "[]"))

    # if (json.parentKey) {
    #   job.parentKey = json.parentKey;
    # }

    # if (json.parent) {
    #   job.parent = JSON.parse(json.parent);
    # }

    return job


Job.fromJSON = staticmethod(fromJSON)


def optsFromJSON(rawOpts: dict) -> dict:
    # opts = json.loads(rawOpts)
    opts = rawOpts

    option_entries = opts.items()

    options = {}
    for item in option_entries:
        attribute_name = item[0]
        value = item[1]
        if attribute_name in optsDecodeMap:
            options[optsDecodeMap[attribute_name]] = value
        else:
            options[attribute_name] = value

    return options


def getReturnValue(value: Any):
    try:
        json.loads(value)
    except Exception as err:
        return value
